"""
State changes for accounts.

Also, the account is the aggregate root for invoices and charges,
so the creation of those is managed here.

"""
from datetime import date
from typing import Dict, Optional, Sequence

from django.db import transaction
from moneyed import Money
from structlog import get_logger

from billing.signals import invoice_ready
from ..models import Account, Charge, Invoice, ProductProperty

logger = get_logger()


def close(account_id: str) -> None:
    """
    Closes the account.

    :param account_id: the account to close
    :return: Nothing
    """
    logger.info('closing-account', account_id=account_id)
    with transaction.atomic():
        account = Account.objects.get(pk=account_id)
        account.close()
        account.save()


def reopen(account_id: str) -> None:
    """
    Reopens the account.

    :param account_id: the account to reopen
    :return: Nothing
    """
    logger.info('reopening-account', account_id=account_id)
    with transaction.atomic():
        account = Account.objects.get(pk=account_id)
        account.reopen()
        account.save()


def create_invoices(account_id: str, due_date: date) -> Sequence[Invoice]:
    """
    Creates the invoices for any uninvoiced charges in the account.
    Creates one invoice per currency (only when the total in that currency is positive).

    :param due_date: The date when the invoices will be due.
    :param account_id: The account to invoice.
    :return: A possibly-empty list of Invoices.
    """
    invoices = []
    with transaction.atomic():
        ucs, total = Charge.objects.uninvoiced_with_total(account_id=account_id)
        for amount_due in total.monies():
            if amount_due.amount > 0:
                invoice = Invoice.objects.create(account_id=account_id, due_date=due_date)
                Charge.objects. \
                    uninvoiced_in_currency(account_id=account_id, currency=amount_due.currency) \
                    .update(invoice=invoice)
                invoices.append(invoice)
    logger.info('created-invoices', account_id=str(account_id), invoice_ids=[i.pk for i in invoices])
    for invoice in invoices:
        invoice_ready.send(sender=create_invoices, invoice=invoice)
    return invoices


def add_charge(account_id: str, amount: Money,
               reverses_id: Optional[str] = None,
               product_code: Optional[str] = None,
               product_properties: Optional[Dict[str, str]] = None) -> Charge:
    """
    Add a charge to the account.

    :param account_id: The account on which to add the charge
    :param amount:  The amount of the charge
    :param product_code: A code identifying the type of product cnarged
    :param product_properties: A dict of hames and values.
    :param ad_hoc_label:
    :return: The newly created charge
    """
    logger.info('adding-charge', account_id=account_id, amount=amount, product_code=product_code,
                product_properties=product_properties)

    with transaction.atomic():
        charge = Charge(account_id=account_id,
                        amount=amount)
        if reverses_id:
            charge.reverses_id = reverses_id
        if product_code:
            charge.product_code = product_code
        charge.full_clean(exclude=['id', 'account'])  # Exclude to avoid unnecessary db queries
        charge.save(force_insert=True)

        if product_properties:
            objs = [ProductProperty(charge=charge, name=k, value=v) for k, v in product_properties.items()]
            for o in objs:
                o.full_clean(exclude=['id', 'charge'])  # Exclude to avoid unnecessary db queries
            ProductProperty.objects.bulk_create(objs)

    return charge



"""
Adapted from django-hordak.


``Total`` **instances (see below for more details)**:

    An account can hold multiple currencies, and a `Total`_ instance is how we represent this.

    A `Total`_ may contain one or more ``Money`` objects. There will be precisely one ``Money`` object
    for each currency which the account holds.

    `Total`_ objects may be added, subtracted etc. This will produce a new `Total`_ object containing a
    union of all the currencies involved in the calculation, even where the result was zero.

"""

import copy
from django.utils.translation import get_language
from django.utils.translation import to_locale
from moneyed import Money
from moneyed.localization import format_money
from rest_framework import serializers
from rest_framework.fields import DecimalField


class Total(object):
    """
    Accounts may have multiple currencies. This class represents these multi-currency
    Totals and provides math functionality. Totals can be added, subtracted,absolute'ed,
    and have their sign changed.

    Examples:

        Example use::

            Total([Money(100, 'USD'), Money(200, 'EUR')])

            # Or in short form
            Total(100, 'USD', 200, 'EUR')
    """

    def __init__(self, _money_obs=None, *args):
        all_args = [_money_obs] + list(args)
        if len(all_args) % 2 == 0:
            _money_obs = []
            for i in range(0, len(all_args) - 1, 2):
                _money_obs.append(Money(all_args[i], all_args[i + 1]))

        self._money_obs = tuple(_money_obs or [])
        self._by_currency = {m.currency.code: m for m in self._money_obs}
        if len(self._by_currency) != len(self._money_obs):
            raise ValueError('Duplicate currency provided. All Money instances must have a unique currency.')

    def __str__(self):
        def fmt(money):
            return format_money(money, locale=to_locale(get_language() or 'en-us'))

        return ', '.join(map(fmt, self._money_obs)) or 'No values'

    def __repr__(self):
        return 'Total: {}'.format(self.__str__())

    def __getitem__(self, currency):
        if hasattr(currency, 'code'):
            currency = currency.code
        elif not isinstance(currency, str) or len(currency) != 3:
            raise ValueError('Currencies must be a string of length three, not {}'.format(currency))

        try:
            return self._by_currency[currency]
        except KeyError:
            return Money(0, currency)

    def __add__(self, other):
        if not isinstance(other, Total):
            raise TypeError('Can only add/subtract Total instances, not Total and {}.'.format(type(other)))
        by_currency = copy.deepcopy(self._by_currency)
        for other_currency, other_money in other._by_currency.items():
            by_currency[other_currency] = other_money + self[other_currency]
        return self.__class__(by_currency.values())

    def __sub__(self, other):
        return self.__add__(-other)

    def __neg__(self):
        return self.__class__([-m for m in self._money_obs])

    def __pos__(self):
        return self.__class__([+m for m in self._money_obs])

    def __abs__(self):
        return self.__class__([abs(m) for m in self._money_obs])

    def __bool__(self):
        return any([bool(m) for m in self._money_obs])

    def __eq__(self, other):
        if other == 0:
            # Support comparing to integer/Decimal zero as it is useful
            return not self.__bool__()
        elif not isinstance(other, Total):
            raise TypeError('Can only compare Total objects to other '
                            'Total objects, not to type {}'.format(type(other)))
        return not self - other

    def __ne__(self, other):
        return not self.__eq__(other)

    def monies(self):
        """Get a list of the underlying ``Money`` instances

        Returns:
            ([Money]): A list of zero or more money instances. Currencies will be unique.
        """
        return [copy.copy(m) for m in self._money_obs]

    def nonzero_monies(self):
        """Get a list of the underlying ``Money`` instances that are not zero

        Returns:
            ([Money]): A list of zero or more money instances. Currencies will be unique.
        """
        return [copy.copy(m) for m in self._money_obs if m.amount != 0]

    def currencies(self):
        """Get all currencies, including those with zero values"""
        return [m.currency.code for m in self.monies() if m.amount]


class TotalSerializer(serializers.BaseSerializer):
    """
    Totals are serialized as a list of money instances.
    """
    amount_serializer = DecimalField(max_digits=12, decimal_places=2)

    def to_representation(self, obj):
        # We cannot use djmoney.contrib.django_rest_framework.MoneyField because a total is not a field.
        # So we replicate the output.
        return [{'amount': TotalSerializer.amount_serializer.to_representation(money.amount),
                 'amount_currency': money.currency.code} for money in obj.nonzero_monies()]



# Generated by Django 2.0.1 on 2018-01-17 14:35

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0005_creditcard_status'),
    ]

    operations = [
        migrations.CreateModel(
            name='ProductProperty',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=100, validators=[django.core.validators.RegexValidator(flags=258, message='a letter maybe followed by letters, numbers, or underscores', regex='^[a-z]\\w*$')])),
                ('value', models.CharField(max_length=255)),
            ],
        ),
        migrations.RemoveField(
            model_name='charge',
            name='description',
        ),
        migrations.AddField(
            model_name='charge',
            name='ad_hoc_label',
            field=models.TextField(blank=True, help_text='When not empty, this is shown verbatim to the user.'),
        ),
        migrations.AddField(
            model_name='charge',
            name='product_code',
            field=models.CharField(blank=True, db_index=True, help_text='Identifies the kind of product being charged or credited', max_length=8, validators=[django.core.validators.RegexValidator(message='Between 4 and 8 uppercase letters or digits', regex='^[A-Z0-9]{4,8}$')]),
        ),
        migrations.AddField(
            model_name='productproperty',
            name='charge',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='product_properties', to='billing.Charge'),
        ),
        migrations.AlterUniqueTogether(
            name='productproperty',
            unique_together={('charge', 'name')},
        ),
    ]



# Generated by Django 2.0.1 on 2018-01-26 10:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0009_longer_product_code'),
    ]

    operations = [
        migrations.AlterField(
            model_name='productproperty',
            name='value',
            field=models.CharField(blank=True, max_length=255),
        ),
    ]



from django.http import Http404
from rest_framework import permissions, serializers
from rest_framework.decorators import permission_classes
from rest_framework.generics import RetrieveAPIView
from rest_framework.mixins import UpdateModelMixin, RetrieveModelMixin, ListModelMixin
from rest_framework.viewsets import GenericViewSet

from .models import Account, Charge, CreditCard, Invoice, Transaction, ProductProperty
from .total import TotalSerializer


class CreditCardSerializer(serializers.ModelSerializer):
    class Meta:
        model = CreditCard
        exclude = ['account', 'expiry_date', 'psp_content_type', 'psp_object_id']


class CreditCardUpdateSerializer(serializers.ModelSerializer):
    class Meta:
        model = CreditCard
        fields = ['status']

    def update(self, instance, validated_data):
        new_status = validated_data['status']
        if new_status == CreditCard.INACTIVE:
            instance.deactivate()
        elif new_status == CreditCard.ACTIVE:
            instance.reactivate()
        else:
            raise Exception('Unknown status')
        instance.save()
        return instance


@permission_classes([permissions.IsAuthenticated])
class CreditCardViewSet(ListModelMixin, RetrieveModelMixin, UpdateModelMixin, GenericViewSet):
    """
    list: Return the list of credit cards registered on the account.
    retrieve: Return the credit card information
    partial_update: Change the status of the creditcard.
    """
    http_method_names = ['get', 'patch']  # We don't want put (inherited from UpdateModelMixin)

    def get_queryset(self):
        return CreditCard.objects.filter(account__owner=self.request.user)

    def get_serializer_class(self):
        method = self.request.method
        if method == 'GET':
            return CreditCardSerializer
        elif method == 'PATCH':
            return CreditCardUpdateSerializer
        else:
            raise Exception('Unknown method')


########################################################################################################

class ProductPropertyListSerializer(serializers.ListSerializer):
    # From: https://stackoverflow.com/questions/31583445
    def to_representation(self, data):
        r = super().to_representation(data)
        return {item['name']: item['value'] for item in r}


class ProductPropertySerializer(serializers.ModelSerializer):
    class Meta:
        model = ProductProperty
        fields = ['name', 'value']
        list_serializer_class = ProductPropertyListSerializer


class ChargeSerializer(serializers.ModelSerializer):
    product_properties = ProductPropertySerializer(read_only=True, many=True)

    class Meta:
        model = Charge
        exclude = ['account', 'deleted', 'reverses']


########################################################################################################

class TransactionSerializer(serializers.ModelSerializer):
    class Meta:
        model = Transaction
        exclude = ['account', 'psp_content_type', 'psp_object_id']


########################################################################################################

class InvoiceSerializer(serializers.ModelSerializer):
    total = TotalSerializer(read_only=True)

    class Meta:
        model = Invoice
        exclude = ['account']


########################################################################################################

class AccountSerializer(serializers.ModelSerializer):
    balance = TotalSerializer(read_only=True)
    credit_cards = CreditCardSerializer(read_only=True, many=True)
    charges = ChargeSerializer(read_only=True, many=True)
    invoices = InvoiceSerializer(read_only=True, many=True)
    transactions = TransactionSerializer(read_only=True, many=True)

    class Meta:
        model = Account
        exclude = ['owner']


@permission_classes([permissions.IsAuthenticated])
class AccountView(RetrieveAPIView):
    """
    get: Retrieve the full billing account information.
    """
    serializer_class = AccountSerializer
    pagination_class = None

    def get_object(self):
        try:
            return Account.objects.open() \
                .prefetch_related('invoices') \
                .prefetch_related('credit_cards') \
                .prefetch_related('transactions') \
                .prefetch_related('charges__product_properties') \
                .get(owner=self.request.user)
        except Account.DoesNotExist:
            raise Http404('No Account matches the given query.')



import logging
from datetime import timedelta, date

import progressbar
import structlog
from collections import defaultdict
from django.core.management.base import BaseCommand
from django.utils import timezone, dateparse

from ...actions.accounts import create_invoices
from ...models import Account


def set_debug(logger_name):
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler())


logger = structlog.get_logger()


def parse_due_date(s):
    d = dateparse.parse_date(s)
    if d is None:
        raise ('Not a valid iso8601 date: {}'.format(s))
    return d


class Command(BaseCommand):
    help = """Create invoices for all the accounts that have pending charges.
              Pass -v 2 to see sql queries.
              """

    def add_arguments(self, parser):
        parser.add_argument('--quiet-days', type=int, required=True,
                            help='Accounts with charges that appeared since quiet-days will not be invoiced. '
                                 '0 means invoice all.')
        parser.add_argument('--due-date', type=parse_due_date, default=date.today(),
                            help='The due date of the invoices that are created. Defaults to today')
        parser.add_argument('--dry-run', action='store_true',
                            help='Counts the accounts that will be invoiced and exits')
        parser.add_argument('--progress', action='store_true', help='Displays a progress bar')

    def handle(self, *args, **options):
        if options['verbosity'] >= 2:
            set_debug('django.db.backends')

        accounts = Account.objects.open().with_uninvoiced_positive_charges()

        quiet_days = options['quiet_days']
        if quiet_days != 0:
            dt = timezone.now() - timedelta(days=quiet_days)
            accounts = accounts.with_no_charges_since(dt)

        dry_run = options['dry_run']
        due_date = options['due_date']

        logger.info('create-invoices-start', dry_run=dry_run, quiet_days=quiet_days, due_date=due_date,
                    invoicable_accounts=len(accounts))

        if dry_run:
            return

        if options['progress']:
            bar = progressbar.ProgressBar()
            accounts = bar(accounts)

        try:
            stats = defaultdict(lambda: 0)
            for account in accounts:
                try:
                    invoices = create_invoices(account_id=account.pk, due_date=due_date)
                    stats_key = '{}_invoices'.format(len(invoices))
                    stats[stats_key] += 1
                except Exception as ex:
                    logger.error('error', account_id=account.pk, ex=ex)
                    stats['error'] += 1
        finally:
            logger.info('create-invoices-done', **stats)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-11-17 15:21
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0003_psp_uri_to_generic_relation'),
    ]

    operations = [
        migrations.AddField(
            model_name='account',
            name='modified',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name='charge',
            name='modified',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name='creditcard',
            name='modified',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name='invoice',
            name='modified',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name='transaction',
            name='modified',
            field=models.DateTimeField(auto_now=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-12-19 17:37
from __future__ import unicode_literals

from django.db import migrations
import django_fsm


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0004_add_modified_field'),
    ]

    operations = [
        migrations.AddField(
            model_name='creditcard',
            name='status',
            field=django_fsm.FSMField(choices=[('ACTIVE', 'Active'), ('INACTIVE', 'Inactive')], db_index=True, default='ACTIVE', max_length=20),
        ),
    ]



# Generated by Django 2.0.1 on 2018-02-02 17:29

from django.db import migrations, models
import django_fsm


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0012_index_deleted_field'),
    ]

    operations = [
        migrations.AddField(
            model_name='invoice',
            name='due_date',
            field=models.DateField(default='2000-01-01'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='invoice',
            name='status',
            field=django_fsm.FSMField(choices=[('PENDING', 'Pending'), ('PAID', 'Paid'), ('CANCELLED', 'Cancelled')], db_index=True, default='PENDING', max_length=20),
        ),
    ]



from typing import Optional

from django.db import transaction
from structlog import get_logger

from .. import psp
from ..models import CreditCard, Invoice, Transaction

logger = get_logger()


class PreconditionError(Exception):
    pass


def pay_with_account_credit_cards(invoice_id) -> Optional[Transaction]:
    """
    Get paid for the invoice, trying the valid credit cards on record for the account.

    If successful attaches the payment to the invoice and marks the invoice as paid.

    :param invoice_id: the id of the invoice to pay.
    :return: A successful transaction, or None if we weren't able to pay the invoice.
    """
    logger.debug('invoice-payment-started', invoice_id=invoice_id)
    with transaction.atomic():
        invoice = Invoice.objects.select_for_update().get(pk=invoice_id)

        # Invoice should be in a state that allows payment
        if not invoice.in_payable_state:
            raise PreconditionError('Cannot pay invoice with status {}.'.format(invoice.status))

        # The invoice needs a positive total, in a single currency
        total = invoice.total().monies()
        if len(total) == 0:
            raise PreconditionError('Cannot pay empty invoice.')
        if len(total) > 1:
            raise PreconditionError('Cannot pay invoice with more than one currency.')
        amount = total[0]
        if amount.amount <= 0:
            raise PreconditionError('Cannot pay invoice with non-positive amount.')

        # Try with valid credit cards. Start with the active ones
        valid_credit_cards = CreditCard.objects.valid().filter(account=invoice.account).order_by('status')
        if not valid_credit_cards:
            raise PreconditionError('No valid credit card on account.')

        for credit_card in valid_credit_cards:
            try:
                success, payment_psp_object = psp.charge_credit_card(
                    credit_card_psp_object=credit_card.psp_object,
                    amount=amount,
                    client_ref=str(invoice_id))
                payment = Transaction.objects.create(
                    account=invoice.account,
                    invoice=invoice,
                    amount=amount,
                    success=success,
                    payment_method=credit_card.type,
                    credit_card_number=credit_card.number,
                    psp_object=payment_psp_object)
                if success:
                    invoice.pay()
                    invoice.save()
                    logger.info('invoice-payment-success', invoice=invoice_id, payment=payment)
                    return payment
                else:
                    logger.info('invoice-payment-failure', invoice=invoice_id, payment=payment)
            except Exception as e:
                logger.error('invoice-payment-error', invoice_id=invoice_id, credit_card=credit_card, exc_info=e)
        return None



from datetime import datetime, date

from django import forms
from django.conf.urls import url
from django.contrib import admin
from django.db.models import Max, Prefetch
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse, NoReverseMatch
from django.utils.html import format_html, format_html_join
from django.utils.translation import ugettext_lazy as _
from moneyed.localization import format_money
from structlog import get_logger

from .actions import accounts, invoices
from .models import Account, Charge, CreditCard, Invoice, Transaction, ProductProperty

logger = get_logger()


##############################################################
# Shared utilities

class AppendOnlyModelAdmin(admin.ModelAdmin):
    """
    Adapted from: https://gist.github.com/aaugustin/1388243
    """

    def get_readonly_fields(self, request, obj=None):
        # Make everything readonly (unless superuser)
        if request.user.is_superuser:
            return super().get_readonly_fields(request, obj)

        # We cannot call super().get_fields(request, obj) because that method calls
        # get_readonly_fields(request, obj), causing infinite recursion. Ditto for
        # super().get_form(request, obj). So we  assume the default ModelForm.

        f = self.fields or [f.name for f in self.model._meta.fields]
        f.extend(super().get_readonly_fields(request, obj))
        return f

    def has_change_permission(self, request, obj=None):
        # Allow viewing objects but not actually changing them (unless superuser)
        if request.user.is_superuser:
            return True
        return request.method in ['GET', 'HEAD'] and super().has_change_permission(request, obj)

    def get_actions(self, request):
        # Disable delete action (unless superuser)
        actions = super().get_actions(request)
        if not request.user.is_superuser:
            if 'delete_selected' in actions:
                del actions['delete_selected']
        return actions

    def has_delete_permission(self, request, obj=None):
        # Disable delete link on detail page (unless superuser)
        return request.user.is_superuser


account_owner_search_fields = ['account__owner__email', 'account__owner__first_name', 'account__owner__last_name']


def amount(obj):
    return format_money(obj.amount)


amount.admin_order_field = 'amount'  # type: ignore


def created_on(obj):
    return obj.created.date()


created_on.admin_order_field = 'created'  # type: ignore
created_on.short_description = 'created'  # type: ignore


def modified_on(obj):
    return obj.modified.date()


modified_on.admin_order_field = 'modified'  # type: ignore
modified_on.short_description = 'modified'  # type: ignore


def psp_admin_link(obj):
    if obj.psp_content_type is None or obj.psp_object_id is None:
        return '-'

    text = '{}: {}'.format(obj.psp_content_type.name, obj.psp_object_id)
    try:
        url = reverse(
            'admin:{}_{}_change'.format(
                obj.psp_content_type.app_label,
                obj.psp_content_type.model),
            args=[obj.psp_object_id])
        return format_html('<a href="{}">{}</a>', url, text)
    except NoReverseMatch:
        return None


psp_admin_link.short_description = 'PSP Object'  # type: ignore


def link_to_account(obj):
    account = obj.account
    text = str(account)
    url = reverse('admin:billing_account_change', args=(account.pk,))
    return format_html('<a href="{}">{}</a>', url, text)


link_to_account.admin_order_field = 'account'  # type: ignore
link_to_account.short_description = 'Account'  # type: ignore


def link_to_invoice(obj):
    invoice_id = obj.invoice_id
    if invoice_id is None:
        return '-'
    else:
        text = '#{}'.format(invoice_id)
        url = reverse('admin:billing_invoice_change', args=(invoice_id,))
        return format_html('<a href="{}">{}</a>', url, text)


link_to_invoice.admin_order_field = 'invoice'  # type: ignore
link_to_invoice.short_description = 'Invoice'  # type: ignore


##############################################################
# Credit Cards

class CreditCardValidFilter(admin.SimpleListFilter):
    title = _('Valid')
    parameter_name = 'valid'

    def lookups(self, request, model_admin):
        return (
            ('yes', _('Yes')),
            ('no', _('No')),
            ('all', _('All')),
        )

    def queryset(self, request, queryset):
        today = datetime.now().date()
        if self.value() == 'yes':
            return queryset.filter(expiry_date__gte=today)
        if self.value() == 'no':
            return queryset.filter(expiry_date__lt=today)


def credit_card_expiry(obj):
    return format_html('{}/{}', obj.expiry_month, obj.expiry_year)


credit_card_expiry.admin_order_field = 'expiry_date'  # type: ignore


def credit_card_is_valid(obj):
    return obj.is_valid()


credit_card_is_valid.boolean = True  # type: ignore

credit_card_is_valid.short_description = 'valid'  # type: ignore


@admin.register(CreditCard)
class CreditCardAdmin(AppendOnlyModelAdmin):
    date_hierarchy = 'created'
    list_display = ['number', created_on, link_to_account, 'type', 'status', credit_card_expiry,
                    credit_card_is_valid,
                    psp_admin_link]
    search_fields = ['number'] + account_owner_search_fields
    list_filter = ['type', 'status', CreditCardValidFilter]
    ordering = ['-created']
    list_select_related = True

    raw_id_fields = ['account']
    readonly_fields = ['created', 'modified', 'expiry_date']


class CreditCardInline(admin.TabularInline):
    model = CreditCard
    fields = readonly_fields = ['type', 'number', 'status', credit_card_expiry, created_on, psp_admin_link]
    show_change_link = True
    can_delete = False
    extra = 0
    ordering = ['-created']


##############################################################
# Charges

def charge_deleted(charge):
    """
    We don't want to display a green check box to mean deleted.
    We use text instead to make sure there is no misinterpretation.
    """
    return 'Yes' if charge.deleted else 'No'


charge_deleted.short_description = 'Deleted'  # type: ignore

charge_deleted.admin_order_field = 'deleted'  # type: ignore


class ProductPropertyInline(admin.TabularInline):
    model = ProductProperty
    verbose_name_plural = 'Product props'
    ordering = ['name']


def product_properties(obj):
    return format_html_join(
        ',\n',
        '<strong>{}</strong>: {}',
        ((p.name, p.value) for p in obj.product_properties.all()))


product_properties.short_description = 'Product props'  # type: ignore


@admin.register(Charge)
class ChargeAdmin(AppendOnlyModelAdmin):
    date_hierarchy = 'created'
    list_display = ['type', created_on, charge_deleted, link_to_account, amount, 'product_code', product_properties,
                    'ad_hoc_label', link_to_invoice]
    search_fields = ['id', 'amount', 'amount_currency', 'product_code', 'product_properties__value', 'ad_hoc_label',
                     'invoice__id'] + account_owner_search_fields
    list_filter = ['deleted', 'amount_currency', 'product_code']
    ordering = ['-created']
    list_select_related = True

    raw_id_fields = ['account', 'invoice', 'reverses']
    readonly_fields = ['created', 'modified']
    inlines = [ProductPropertyInline]

    def get_queryset(self, request):
        # In this particular admin screen we want to see even the deleted charges
        qs = Charge.all_charges
        ordering = self.get_ordering(request)
        if ordering:
            qs = qs.order_by(*ordering)
        return qs \
            .prefetch_related('invoice') \
            .prefetch_related('product_properties')


class ChargeInline(admin.TabularInline):
    verbose_name_plural = 'Charges and Credits'
    model = Charge
    fields = readonly_fields = ['type', created_on, link_to_invoice, amount, 'product_code', product_properties,
                                'ad_hoc_label']
    show_change_link = True
    can_delete = False
    extra = 0
    ordering = ['-created']

    def get_queryset(self, request):
        return super().get_queryset(request).prefetch_related('product_properties')


#############################################################
# Transactions

@admin.register(Transaction)
class TransactionAdmin(AppendOnlyModelAdmin):
    verbose_name_plural = 'Transactions'
    date_hierarchy = 'created'
    list_display = ['type', created_on, link_to_account, 'payment_method', 'credit_card_number', 'success',
                    link_to_invoice, amount, psp_admin_link]
    list_display_links = ['type']
    search_fields = ['credit_card_number', 'amount'] + account_owner_search_fields
    list_filter = ['payment_method', 'success', 'amount_currency']
    ordering = ['-created']
    list_select_related = True

    raw_id_fields = ['account', 'invoice']
    readonly_fields = ['created', 'modified']


class TransactionInline(admin.TabularInline):
    model = Transaction
    fields = readonly_fields = ['type', 'payment_method', 'credit_card_number', created_on, 'success', amount,
                                psp_admin_link]
    show_change_link = True
    can_delete = False
    extra = 0
    ordering = ['-created']


##############################################################
# Invoices

class InvoiceOverdueFilter(admin.SimpleListFilter):
    title = _('Overdue')
    parameter_name = 'overdue'

    def lookups(self, request, model_admin):
        return (
            ('yes', _('Yes')),
            ('no', _('No')),
            ('all', _('All')),
        )

    def queryset(self, request, queryset):
        today = datetime.now().date()
        if self.value() == 'yes':
            return queryset.filter(due_date__lt=today)
        if self.value() == 'no':
            return queryset.filter(due_date__gte=today)


class InvoiceValidCCFilter(admin.SimpleListFilter):
    title = _('Valid cc')
    parameter_name = 'valid_cc'

    def lookups(self, request, model_admin):
        return (
            ('yes', _('Yes')),
            ('no', _('No')),
            ('all', _('All')),
        )

    def queryset(self, request, queryset):
        d = date.today()
        if self.value() == 'yes':
            return queryset.filter(account__credit_cards__expiry_date__gte=d)
        if self.value() == 'no':
            return queryset.exclude(account__credit_cards__expiry_date__gte=d)


def pay_invoice_button(obj):
    if obj.pk and obj.in_payable_state:
        return format_html(
            '<a href="{}">Pay Now</a>',
            reverse('admin:billing-pay-invoice', args=[obj.pk]))
    else:
        return '-'


pay_invoice_button.short_description = _('Pay Invoice')  # type: ignore


def do_pay_invoice(request, invoice_id):
    invoices.pay_with_account_credit_cards(invoice_id)
    return HttpResponseRedirect(request.META.get('HTTP_REFERER'))


def invoice_number(invoice):
    return str(invoice)


invoice_number.short_description = 'Invoice'  # type: ignore

invoice_number.admin_order_field = 'pk'  # type: ignore


def invoice_last_transaction(obj):
    dt = obj.last_transaction
    if dt:
        return dt.date()


invoice_last_transaction.short_description = 'Last transaction'  # type: ignore


def invoice_account_has_valid_cc(obj):
    return len(obj.account.valid_credit_card_ids) != 0


invoice_account_has_valid_cc.short_description = 'Valid cc'  # type: ignore

invoice_account_has_valid_cc.boolean = True  # type: ignore


@admin.register(Invoice)
class InvoiceAdmin(AppendOnlyModelAdmin):
    date_hierarchy = 'created'
    list_display = [invoice_number, created_on, link_to_account, invoice_account_has_valid_cc, 'total',
                    'due_date', invoice_last_transaction, 'status']
    list_filter = [InvoiceValidCCFilter, InvoiceOverdueFilter, 'status']
    search_fields = ['id', 'account__owner__email', 'account__owner__first_name', 'account__owner__last_name']
    ordering = ['-created']

    raw_id_fields = ['account']
    readonly_fields = ['created', 'modified', 'total', pay_invoice_button]
    inlines = [ChargeInline, TransactionInline]

    def get_queryset(self, request):
        return super().get_queryset(request) \
            .select_related('account__owner') \
            .annotate(last_transaction=Max('transactions__created')) \
            .prefetch_related(Prefetch('account__credit_cards',
                                       queryset=CreditCard.objects.valid(),
                                       to_attr='valid_credit_card_ids'))

    def get_urls(self):
        custom_urls = [
            url(r'^(?P<invoice_id>[0-9a-f-]+)/pay/$',
                self.admin_site.admin_view(do_pay_invoice),
                name='billing-pay-invoice'),
        ]
        return custom_urls + super().get_urls()


class InvoiceInline(admin.TabularInline):
    model = Invoice
    readonly_fields = [invoice_number, created_on, 'status', 'due_date', 'total']
    show_change_link = True
    can_delete = False
    extra = 0
    ordering = ['-created']


##############################################################
# Accounts

class AccountHasValidCCFilter(admin.SimpleListFilter):
    title = _('Has valid cc')
    parameter_name = 'Has_valid_cc'

    def lookups(self, request, model_admin):
        return (
            ('yes', _('Yes')),
            ('no', _('No')),
            ('all', _('All')),
        )

    def queryset(self, request, queryset):
        d = date.today()
        if self.value() == 'yes':
            return queryset.filter(credit_cards__expiry_date__gte=d)
        if self.value() == 'no':
            return queryset.exclude(credit_cards__expiry_date__gte=d)


def payable_invoice_count(obj):
    return len(obj.payable_invoice_ids)


def has_valid_cc(obj):
    return len(obj.valid_credit_card_ids) != 0


has_valid_cc.boolean = True  # type: ignore


def create_invoices_button(obj):
    if obj.pk:
        return format_html(
            '<a class="button" href="{}">Create Invoices</a>',
            reverse('admin:billing-create-invoices', args=[obj.pk]),
        )
    else:
        return '-'


create_invoices_button.short_description = _('Create Invoices')  # type: ignore


class CreateInvoicesForm(forms.Form):
    due_date = forms.DateField()


def create_invoices_form(request, account_id):
    form = CreateInvoicesForm(request.POST or None, initial={'due_date': date.today()})
    if request.method == 'POST':
        if form.is_valid():
            accounts.create_invoices(
                account_id=account_id,
                due_date=form.cleaned_data['due_date'])
            # As confirmation take the user to the account overview.
            return HttpResponseRedirect(reverse('admin:billing_account_change', args=[account_id]))

    return render(
        request,
        'admin/billing/form.html',
        {
            'title': 'Create invoices',
            'form': form,
            'opts': Account._meta,  # Used to setup the navigation / breadcrumbs of the page
        }
    )


def do_create_invoices(request, account_id):
    accounts.create_invoices(account_id=account_id)
    return HttpResponseRedirect(request.META.get('HTTP_REFERER'))


@admin.register(Account)
class AccountAdmin(AppendOnlyModelAdmin):
    date_hierarchy = 'created'
    list_display = ['owner', created_on, modified_on, payable_invoice_count, has_valid_cc, 'currency', 'status']
    search_fields = ['id', 'owner__email', 'owner__first_name', 'owner__last_name']
    ordering = ['-created']
    list_filter = [AccountHasValidCCFilter, 'currency', 'status']
    list_select_related = True

    raw_id_fields = ['owner']
    readonly_fields = ['balance', 'created', 'modified', create_invoices_button]

    inlines = [CreditCardInline, ChargeInline, InvoiceInline, TransactionInline]

    def get_urls(self):
        urls = super().get_urls()
        my_urls = [
            url(
                r'^(?P<account_id>[0-9a-f-]+)/create_invoices/$',
                self.admin_site.admin_view(create_invoices_form),
                name='billing-create-invoices'
            )
        ]
        return my_urls + urls

    def get_queryset(self, request):
        return super().get_queryset(request) \
            .prefetch_related(Prefetch('invoices',
                                       queryset=Invoice.objects.payable().only('id'),
                                       to_attr='payable_invoice_ids')) \
            .prefetch_related(Prefetch('credit_cards',
                                       queryset=CreditCard.objects.valid().only('id'),
                                       to_attr='valid_credit_card_ids'))



# Generated by Django 2.0.1 on 2018-01-27 06:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0011_cascade_delete_product_properties'),
    ]

    operations = [
        migrations.AlterField(
            model_name='charge',
            name='deleted',
            field=models.BooleanField(db_index=True, default=False),
        ),
    ]



from django.db import transaction
from structlog import get_logger

from .accounts import add_charge
from ..models import Charge

logger = get_logger()

REVERSAL_PRODUCT_CODE = 'REVERSAL'


class ChargeAlreadyCancelledError(Exception):
    pass


def cancel_charge(charge_id: str) -> None:
    """
    Cancels an existing charge.

    If the charge was already cancelled then an Exception is raised.

    If it is not in an invoice then the charge is deleted,
    otherwise a Credit object is created to reverse the Charge.

    :param charge_id: The id of the charge to cancel.
    """
    logger.info('cancelling-charge', charge_id=charge_id)

    with transaction.atomic():
        charge = Charge.all_charges.get(pk=charge_id)

        if charge.deleted:
            raise ChargeAlreadyCancelledError("Cannot cancel deleted charge.")

        if Charge.all_charges.filter(reverses=charge_id).exists():
            raise ChargeAlreadyCancelledError("Cannot cancel reversed charge.")

        if charge.invoice is None:
            charge.deleted = True
            charge.save()
        else:
            add_charge(
                account_id=charge.account_id,
                reverses_id=charge_id,
                amount=-charge.amount,
                product_code=REVERSAL_PRODUCT_CODE)



# Generated by Django 2.0.2 on 2018-02-10 13:26

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0013_replace_past_due_status_by_due_date'),
    ]

    operations = [
        migrations.AlterField(
            model_name='charge',
            name='invoice',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='items', to='billing.Invoice'),
        ),
    ]



import logging

import progressbar
import structlog
from collections import defaultdict
from django.core.management.base import BaseCommand

from ...actions.invoices import pay_with_account_credit_cards
from ...models import Invoice, CreditCard


def set_debug(logger_name):
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler())


logger = structlog.get_logger()


class Command(BaseCommand):
    help = """Pay pending invoices with credit cards registered on accounts.
              Pass v2 to see sql queries"""

    def add_arguments(self, parser):
        parser.add_argument('--dry-run', action='store_true', dest='dry_run',
                            help="Displays the payable invoices but doesn't perform any action.")
        parser.add_argument('--progress', action='store_true', dest='progress',
                            help='Displays a progress bar')

    def handle(self, *args, **options):
        if options['verbosity'] >= 2:
            set_debug('django.db.backends')

        dry_run = options['dry_run']

        all_payable_invoices = Invoice.objects.payable()

        logger.debug('pay-invoice-select', dry_run=dry_run, payable=len(all_payable_invoices))

        # Should replace by a filter, to run in a single sql query
        invoices = [
            i for i in all_payable_invoices if CreditCard.objects.valid().filter(account_id=i.account_id).exists()
        ]

        logger.info('pay-invoices-start', dry_run=dry_run, payable_with_valid_cc=len(invoices))

        if dry_run:
            return

        if options['progress']:
            bar = progressbar.ProgressBar()
            invoices = bar(invoices)

        try:
            stats = defaultdict(lambda: 0)
            for invoice in invoices:
                try:
                    maybe_transaction = pay_with_account_credit_cards(invoice.pk)
                    if maybe_transaction is not None:
                        stats['success'] += 1
                    else:
                        stats['failure'] += 1
                except Exception as ex:
                    logger.error('pay-invoices-error', invoice_id=invoice.pk, ex=ex)
                    stats['error'] += 1
        finally:
            logger.info('pay-invoices-done', **stats)



from django.db import transaction
from structlog import get_logger

from ..models import CreditCard

logger = get_logger()


def deactivate(credit_card_id: str) -> None:
    """
    Deactivates a credit card.
    """
    logger.info('deactivating-credit-card', credit_card_id=credit_card_id)
    with transaction.atomic():
        cc = CreditCard.objects.get(pk=credit_card_id)
        cc.deactivate()
        cc.save()


def reactivate(credit_card_id: str) -> None:
    """
    Reactivates a credit card.
    """
    logger.info('reactivating-credit-card', credit_card_id=credit_card_id)
    with transaction.atomic():
        cc = CreditCard.objects.get(pk=credit_card_id)
        cc.reactivate()
        cc.save()



from django.conf.urls import url
from rest_framework import routers

from .views import AccountView, CreditCardViewSet

router = routers.SimpleRouter(trailing_slash=False)
router.register(r'account/credit-cards', CreditCardViewSet, basename='billing_creditcard')

urlpatterns = router.urls + [
    url(r'^account$', AccountView.as_view(), name='billing_account'),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-11-10 08:46
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('billing', '0002_default_currency'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='creditcard',
            name='psp_uri',
        ),
        migrations.RemoveField(
            model_name='transaction',
            name='psp_uri',
        ),
        migrations.AddField(
            model_name='creditcard',
            name='psp_content_type',
            field=models.ForeignKey(default=1, on_delete=django.db.models.deletion.CASCADE,
                                    to='contenttypes.ContentType'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='creditcard',
            name='psp_object_id',
            field=models.UUIDField(db_index=True, default='00000000-0000-0000-0000-000000000000'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='transaction',
            name='psp_content_type',
            field=models.ForeignKey(default=1, on_delete=django.db.models.deletion.CASCADE,
                                    to='contenttypes.ContentType'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='transaction',
            name='psp_object_id',
            field=models.UUIDField(db_index=True, default='00000000-0000-0000-0000-000000000000'),
            preserve_default=False,
        ),
    ]



# Generated by Django 2.0.1 on 2018-01-20 17:35

from django.db import migrations
import django_fsm


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0006_add_product_code_and_properties'),
    ]

    operations = [
        migrations.AlterField(
            model_name='invoice',
            name='status',
            field=django_fsm.FSMField(choices=[('PENDING', 'Pending'), ('PAST_DUE', 'Past-due'), ('PAID', 'Paid'), ('CANCELLED', 'Cancelled')], db_index=True, default='PENDING', max_length=20),
        ),
    ]



import calendar
import re
import uuid
from datetime import date, datetime

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator
from django.core.validators import RegexValidator
from django.db import models
from django.db.models import Model, PROTECT, CASCADE, QuerySet, Sum, SET_NULL
from django.utils.translation import ugettext_lazy as _
from django_fsm import FSMField, can_proceed, transition
from djmoney.models.fields import CurrencyField, MoneyField
from moneyed import Money
from typing import List, Tuple

from .total import Total


def total_amount(qs) -> Total:
    """Sums the amounts of the objects in the queryset, keeping each currency separate.
    :param qs: A querystring containing objects that have an amount field of type Money.
    :return: A Total object.
    """
    aggregate = qs.values('amount_currency').annotate(sum=Sum('amount'))
    return Total(Money(amount=r['sum'], currency=r['amount_currency']) for r in aggregate)


########################################################################################################
# Accounts

class AccountQuerySet(models.QuerySet):
    def open(self):
        return self.filter(status=Account.OPEN)

    def with_uninvoiced_positive_charges(self):
        return self.filter(charges__deleted=False, charges__amount__gt=0, charges__invoice__isnull=True).distinct()

    def with_no_charges_since(self, dt: datetime):
        return self.exclude(charges__created__gte=dt)


class Account(Model):
    OPEN = 'OPEN'
    CLOSED = 'CLOSED'
    STATUS_CHOICES = (
        (OPEN, _('Open')),
        (CLOSED, _('Closed')),
    )
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    created = models.DateTimeField(auto_now_add=True, db_index=True)
    modified = models.DateTimeField(auto_now=True)
    owner = models.OneToOneField(settings.AUTH_USER_MODEL, related_name='billing_account', on_delete=PROTECT)
    currency = CurrencyField(db_index=True)
    status = FSMField(max_length=20, choices=STATUS_CHOICES, default=OPEN, db_index=True)

    objects = AccountQuerySet.as_manager()

    def balance(self, as_of: date = None):
        charges = Charge.objects.filter(account=self)
        transactions = Transaction.successful.filter(account=self)
        if as_of is not None:
            charges = charges.filter(created__lte=as_of)
            transactions = transactions.filter(created__lte=as_of)
        return total_amount(transactions) - total_amount(charges)

    @transition(field=status, source=OPEN, target=CLOSED)
    def close(self):
        pass

    @transition(field=status, source=CLOSED, target=OPEN)
    def reopen(self):
        pass

    def __str__(self):
        return str(self.owner)


########################################################################################################
# Invoices

class InvoiceQuerySet(models.QuerySet):
    def payable(self, as_of: date = None) -> QuerySet:
        if as_of is None:
            as_of = date.today()
        return self.filter(status=Invoice.PENDING, due_date__lte=as_of)


class Invoice(Model):
    PENDING = 'PENDING'
    PAID = 'PAID'
    CANCELLED = 'CANCELLED'
    STATUS_CHOICES = (
        (PENDING, _('Pending')),
        (PAID, _('Paid')),
        (CANCELLED, _('Cancelled')),
    )
    account = models.ForeignKey(Account, related_name='invoices', on_delete=PROTECT)
    created = models.DateTimeField(auto_now_add=True, db_index=True)
    modified = models.DateTimeField(auto_now=True)
    due_date = models.DateField()
    status = FSMField(max_length=20, choices=STATUS_CHOICES, default=PENDING, db_index=True)

    objects = InvoiceQuerySet.as_manager()

    @transition(field=status, source=[PENDING], target=PAID)
    def pay(self):
        pass

    @transition(field=status, source=[PENDING], target=CANCELLED)
    def cancel(self):
        pass

    @property
    def in_payable_state(self):
        return can_proceed(self.pay)

    def total(self):
        return total_amount(Charge.objects.filter(invoice=self))

    def __str__(self):
        return '#{}'.format(self.id)


########################################################################################################
# Charges

product_code_validator = RegexValidator(regex=r'^[A-Z0-9]{4,10}$',
                                        message='Between 4 and 10 uppercase letters or digits')


class ChargeManager(models.Manager):

    def get_queryset(self):
        return super().get_queryset().exclude(deleted=True)

    def uninvoiced_with_total(self, account_id: str) -> Tuple[List, Total]:
        uc = Charge.objects.filter(invoice=None, account_id=account_id)
        return list(uc), total_amount(uc)

    def uninvoiced_in_currency(self, account_id: str, currency: str) -> QuerySet:
        return Charge.objects.filter(invoice=None, account_id=account_id, amount_currency=currency)


class Charge(Model):
    """
    A charge has a signed amount. If the amount is negative then the charge is in fact a credit.
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    created = models.DateTimeField(auto_now_add=True, db_index=True)
    modified = models.DateTimeField(auto_now=True)
    account = models.ForeignKey(Account, on_delete=PROTECT, related_name='charges')
    invoice = models.ForeignKey(Invoice, null=True, blank=True, related_name='items', on_delete=SET_NULL)
    amount = MoneyField(max_digits=12, decimal_places=2)
    ad_hoc_label = models.TextField(blank=True, help_text='When not empty, this is shown verbatim to the user.')
    product_code = models.CharField(max_length=10, blank=True, validators=[product_code_validator], db_index=True,
                                    help_text='Identifies the kind of product being charged or credited')
    reverses = models.OneToOneField('self', null=True, blank=True, related_name='reversed_by', on_delete=PROTECT)
    deleted = models.BooleanField(default=False, db_index=True)

    objects = ChargeManager()

    all_charges = models.Manager()  # Includes deleted charges

    def clean(self):
        if not (self.ad_hoc_label or self.product_code):
            raise ValidationError('Either the ad-hoc-label or the product-code must be filled.')

    @property
    def type(self):
        a = self.amount.amount
        if a >= 0:
            return _('Charge')
        else:
            return _('Credit')


product_property_name_validator = RegexValidator(regex=r'^[a-z]\w*$',
                                                 flags=re.ASCII | re.IGNORECASE,
                                                 message='a letter maybe followed by letters, numbers, or underscores')


class ProductProperty(Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    created = models.DateTimeField(auto_now_add=True, db_index=True)
    modified = models.DateTimeField(auto_now=True)
    charge = models.ForeignKey(Charge, on_delete=CASCADE, related_name='product_properties')
    name = models.CharField(max_length=100, validators=[product_property_name_validator])
    value = models.CharField(max_length=255, blank=True)

    class Meta:
        unique_together = ['charge', 'name']


########################################################################################################
# Transactions

class OnlySuccessfulTransactionsManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(success=True)


class Transaction(Model):
    """
    A transaction has a signed amount. If the amount is positive then it's a payment,
    otherwise it's a refund.
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    created = models.DateTimeField(auto_now_add=True, db_index=True)
    modified = models.DateTimeField(auto_now=True)
    account = models.ForeignKey(Account, related_name='transactions', on_delete=PROTECT)
    success = models.BooleanField()
    invoice = models.ForeignKey(Invoice, related_name='transactions', null=True, blank=True, on_delete=PROTECT)
    amount = MoneyField(max_digits=12, decimal_places=2)
    payment_method = models.CharField(db_index=True, max_length=3)
    credit_card_number = models.CharField(max_length=255, blank=True)

    psp_content_type = models.ForeignKey(ContentType, blank=True, null=True, on_delete=CASCADE)
    psp_object_id = models.UUIDField(blank=True, null=True, db_index=True)
    psp_object = GenericForeignKey('psp_content_type', 'psp_object_id')

    objects = models.Manager()
    successful = OnlySuccessfulTransactionsManager()

    @property
    def type(self):
        a = self.amount.amount
        if a > 0:
            return _('Payment')
        elif a < 0:
            return _('Refund')

    def __str__(self):
        return '{}-{} ({})'.format(
            self.type,
            self.credit_card_number,
            'success' if self.success else 'failure')


########################################################################################################
# Credit Cards

def compute_expiry_date(two_digit_year: int, month: int) -> date:
    year = 2000 + two_digit_year
    _, last_day_of_month = calendar.monthrange(year, month)
    return date(year=year, month=month, day=last_day_of_month)


class CreditCardQuerySet(models.QuerySet):
    def valid(self, as_of: date = None):
        if as_of is None:
            as_of = date.today()
        return self.filter(expiry_date__gte=as_of)


class CreditCard(Model):
    ACTIVE = 'ACTIVE'
    INACTIVE = 'INACTIVE'
    STATUS_CHOICES = (
        (ACTIVE, _('Active')),
        (INACTIVE, _('Inactive')),
    )

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    created = models.DateTimeField(auto_now_add=True, db_index=True)
    modified = models.DateTimeField(auto_now=True)
    account = models.ForeignKey(Account, related_name='credit_cards', on_delete=PROTECT)
    type = models.CharField(db_index=True, max_length=3)
    number = models.CharField(max_length=255)
    expiry_month = models.IntegerField(validators=[MinValueValidator(1), MaxValueValidator(12)])
    expiry_year = models.IntegerField(validators=[MinValueValidator(0), MaxValueValidator(99)])
    expiry_date = models.DateField()  # A field in the database so we can search for expired cards

    psp_content_type = models.ForeignKey(ContentType, on_delete=CASCADE)
    psp_object_id = models.UUIDField(db_index=True)
    psp_object = GenericForeignKey('psp_content_type', 'psp_object_id')

    status = FSMField(max_length=20, choices=STATUS_CHOICES, default=ACTIVE, db_index=True)

    objects = CreditCardQuerySet.as_manager()

    @transition(field=status, source=ACTIVE, target=INACTIVE)
    def deactivate(self):
        pass

    @transition(field=status, source=INACTIVE, target=ACTIVE)
    def reactivate(self):
        pass

    def is_valid(self, as_of: date = None):
        if as_of is None:
            as_of = datetime.now().date()
        return self.expiry_date >= as_of

    def save(self, *args, **kwargs):
        if self.expiry_year is not None and self.expiry_month is not None:
            self.expiry_date = compute_expiry_date(two_digit_year=self.expiry_year, month=self.expiry_month)
        super().save(*args, **kwargs)



# Generated by Django 2.0.1 on 2018-01-26 17:21

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0010_allow_empty_product_property_value'),
    ]

    operations = [
        migrations.AlterField(
            model_name='productproperty',
            name='charge',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='product_properties', to='billing.Charge'),
        ),
    ]



# Generated by Django 2.1.4 on 2018-12-23 15:07

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0014_deleting_invoice_detaches_charges'),
    ]

    operations = [
        migrations.AlterField(
            model_name='transaction',
            name='psp_content_type',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType'),
        ),
        migrations.AlterField(
            model_name='transaction',
            name='psp_object_id',
            field=models.UUIDField(blank=True, db_index=True, null=True),
        ),
    ]



# Generated by Django 2.0.1 on 2018-01-21 13:34

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0007_payed_to_paid'),
    ]

    operations = [
        migrations.AddField(
            model_name='charge',
            name='deleted',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='charge',
            name='reverses',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='reversed_by', to='billing.Charge'),
        ),
    ]



from django.dispatch import Signal

# We want to signal when an invoice was created and is ready to be looked at.
# (we cannot use the built-in post_save signal for that because just after an
# invoice is saved it doesn't yet have charges attached)
invoice_ready = Signal(providing_args=['invoice'], use_caching=True)



#!/usr/bin/env python
from setuptools import setup

import billing

setup(
    name='django-customer-billing',
    version=billing.__version__,
    description='',
    long_description='',
    author='Nicholas Wolff',
    author_email='nwolff@gmail.com',
    url=billing.__URL__,
    download_url='https://pypi.python.org/pypi/django-customer-billing',
    packages=[
        'billing',
        'billing.actions',
        'billing.migrations',
        'billing.management.commands',
        'billing.signals',
    ],
    package_data={'billing': [
        'templates/admin/billing/*.html',
    ]},
    install_requires=[
        'Django>=2.0',
        'django-money',
        'django-fsm',
        'djangorestframework',
        'structlog',
        'typing',
        'progressbar2'
    ],
    license=billing.__licence__,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



from django.db.models import Model
from moneyed import Money
from structlog import get_logger
from typing import List, Tuple

logger = get_logger()


#############################################
# The Client API

class PreconditionError(Exception):
    pass


def charge_credit_card(credit_card_psp_object: Model, amount: Money, client_ref: str) -> Tuple[bool, Model]:
    """
    :param credit_card_psp_object: an instance representing the credit card in the psp
    :param amount: the amount to charge
    :param client_ref: a reference that will appear on the customer's credit card report
    :return: a tuple (success, payment_psp_object)
    """
    logger.debug('charge-credit-card', credit_card_psp_object=credit_card_psp_object, amount=amount,
                 client_ref=client_ref)
    if amount.amount <= 0:
        raise PreconditionError('Can only charge positive amounts')
    psp = psp_for_model_instance(credit_card_psp_object)
    return psp.charge_credit_card(credit_card_psp_object, amount, client_ref)


def refund_payment(payment_psp_object: Model, amount: Money, client_ref: str) -> Tuple[bool, Model]:
    """
    :param payment_psp_object: an instance representing the original payment in the psp
    :param amount: the amount to refund
    :param client_ref: a reference that will appear on the customer's credit card statement
    :return: a tuple (success, refund_psp_object)
    """
    logger.debug('refund-payment', payment_psp_object=payment_psp_object, amount=amount, client_ref=client_ref)
    if amount.amount <= 0:
        raise PreconditionError('Can only refund positive amounts')
    psp = psp_for_model_instance(payment_psp_object)
    return psp.refund_payment(payment_psp_object, amount, client_ref)


#############################################
# The SPI


class PSP:
    """ Each PSP must implement these methods. """

    def model_classes(self) -> List[Model]:
        """
        :return: the list of model classes that this psp is responsible for.
        """
        pass

    def charge_credit_card(self, credit_card_psp_object: Model, amount: Money, client_ref: str) -> Tuple[bool, Model]:
        """
        :param credit_card_psp_object: an instance representing the credit card in the psp
        :param amount: the amount to charge
        :param client_ref: a reference that will appear on the customer's credit card report
        :return: a tuple (success, payment_psp_object)
        """
        pass

    def refund_payment(self, payment_psp_object: Model, amount: Money, client_ref: str) -> Tuple[bool, Model]:
        """
        :param payment_psp_object: an instance representing the original payment in the psp
        :param amount: the amount to refund
        :param client_ref: a reference that will appear on the customer's credit card statement
        :return: a tuple (success, refund_psp_object)
        """
        pass

    class Meta:
        abstract = True


#############################################
# The registry

def register(psp: PSP) -> None:
    for model_class in psp.model_classes():
        _registry[model_class] = psp


def unregister(psp: PSP) -> None:
    for model_class in psp.model_classes():
        del _registry[model_class]


def psp_for_model_instance(model_instance: Model) -> PSP:
    model_class = type(model_instance)
    psp = _registry.get(model_class, None)
    if not psp:
        raise Exception("No PSP registered for model class '{}'".format(model_class))
    return psp


_registry = {}  # type: ignore



# Generated by Django 2.0.1 on 2018-01-26 09:41

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('billing', '0008_charge_deleted_and_reverses'),
    ]

    operations = [
        migrations.AlterField(
            model_name='charge',
            name='product_code',
            field=models.CharField(blank=True, db_index=True, help_text='Identifies the kind of product being charged or credited', max_length=10, validators=[django.core.validators.RegexValidator(message='Between 4 and 10 uppercase letters or digits', regex='^[A-Z0-9]{4,10}$')]),
        ),
    ]
